/*
 * Portable Perl launcher for Windows
 *
 * Approach: Run the Perl interpreter from perl5xx.dll and:
 * - Insert a script name as parameter
 * - Optionally insert include directories as parameter(s)
 *
 * The PAR Packager pp can be used to collect all dependencies of the Perl application
 * Use only the contents of the "lib" directory plus the script from the "script" directory.
 *
 * Inspired by "runperl.c" used to compile the original perl.exe
 *
 * If configured for "implicit linking", the exe needs to
 * be linked with "lib/CORE/libperl5xx.a" from your Perl distribution
 * an perl5xx.dll needs to be placed in the same directory as the exe
 *
 * 20190516 - proof of concept, request for comments
 * 20190518 - catch strrchr not found, add "perl.exe mode"
 * 20190519 - lint-safe
 * 20190521 - explicit linking to build an universal exe (optional)
 * 20190523 - support subdirectory for DLL
 * 20190603 - support for Perl w/ many DLLs (Strawberry Perl)
 * 20190624 - _CRT_glob didn't work on MinGW_i686-8.1.0-release-posix-dwarf-rt_v6-rev0
 * 20190702 - Unicode additions from John Rellis
 *
 * Possible extensions:
 * - set additional lib search path(s)
 * - set an environment variable to let the script know the calling method
 *
 * Author: Oliver Betz https://OliverBetz.de
 * License: Public domain, no liability, no support (useful feedback welcome)
 * Please send comments to ppl@oliverbetz.de
 *
 * Known limitations: Do not put more than one perl5*.dll next to the exe
 * because the FindFirstFile() result would be ambiguous then
 *
 * Unicode compatibility:
 * https://www.nu42.com/2017/02/perl-unicode-windows-trilogy-one.html
 * https://circulosmeos.wordpress.com/2018/11/12/perl-cmdline-with-unicode-in-windows/
 * https://github.com/circulosmeos/Perl-with-Unicode-for-Windows
 *
 */

#define CASSERT(pred) switch(0){case 0:case (pred):;} // compile time assert

#ifdef _MSC_VER
#include <crtdbg.h> // Copied from runperl.c without further check
#endif

#ifdef __GNUC__
int _CRT_glob = 0; // Why the hell should a compiler glob command line arguments?
int _dowildcard = 0; // _CRT_glob should be an alias for _dowildcard but doesn't work
#endif

//#define EXPLICITLINKING ///< load Perl*.dll dynamically (set by makefile)

#ifdef EXPLICITLINKING
#include <stdio.h>
#include <windows.h>
#else
// includes from your Perl distribution to access the DLL by implicit linking
#include "EXTERN.h"
#include "perl.h"
#endif

// Three flavours of mangling the path to the exe can be chosen
// by the format of the pathreplace[] string:
// - "\\myscript.pl" makes a fixed script name, independent of the exe name
// - ".pl" replaces just the extension from .exe to .pl
// - "." removes the .exe extension
// The latter two make an "universal" exe (rename it to match the script name)
// No safety belts - don't omit "\\" or "." at the beginning of the string!

#include <TCHAR.h> // _T() and _ftprintf
#include <fcntl.h> // _O_U16TEXT only needed for DEBUGOUT

//static const char pathreplace[] = ".pl"; ///< variable path conversion to Perl script
static const TCHAR pathreplace[] = _T("\\exiftool_files\\exiftool.pl"); ///< fixed path conversion to Perl script

//static const char dllsearch[] = "perl5*.dll"; ///< search string for Perl DLL
static const TCHAR dllsearch[] = _T("exiftool_files\\perl5*.dll"); ///< search string for Perl DLL

// act like the original "perl.exe" if our name is "perl.exe" (case sensitive!)
static const TCHAR perlexe[] = _T("perl.exe"); // set to NULL to disable

#define ARGS_ADDED 1 ///< number of arguments we put in front of the user provided args
#define DEBUGOUT stderr  ///< debug output. Comment to disable debug output.

#define PATHBUFLEN 1000 ///< PATH_MAX is obsolete since Win10?
#define NAMEBUFLEN   50 ///< additional buffer size for the dll name

// original definition: int RunPerl(int argc, char **argv, char **env);
typedef int (CALLBACK* RunPerl_t)(int argc, char **argv, char **env);

#ifdef UNICODE
// Provided by John Rellis, inspired by https://www.nu42.com/2017/02/perl-unicode-windows-trilogy-one.html
// no free after malloc but this does (probably) no harm, Windows will take care about the memory on exit
static char *
encode(const TCHAR *src, unsigned int codepage)
{
    int len = WideCharToMultiByte(codepage, 0, src, -1, NULL, 0, NULL, NULL);
    char *encoded = malloc(len + 1);
    if (!encoded) {
        _ftprintf(stderr, _T("Failed to allocate memory for encoded string"));
        exit(1);
    }
    (void) WideCharToMultiByte(codepage, 0, src, -1, encoded, len, NULL, NULL);
    return encoded;
}
#endif

int _tmain(int argc, TCHAR **argv, TCHAR **env)
{
    TCHAR scriptpath[PATHBUFLEN]; ///< to construct script path from exe path
    TCHAR dllpath[PATHBUFLEN+NAMEBUFLEN]; ///< to construct DLL (search) path from exe path
    TCHAR *dlldir; // pointer past (!) last backslash in dllpath buffer
    int i, emulate_perlexe;

#if defined UNICODE && defined DEBUGOUT
    _setmode(_fileno(DEBUGOUT), (DEBUGOUT == stdout) ? _O_U16TEXT : _O_U8TEXT); // else wprintf doesn't print UTF-16 characters. Breaks fprint!?
#endif

// Copied from runperl.c without further check:
#ifdef _MSC_VER
    int currentFlag = _CrtSetDbgFlag(_CRTDBG_REPORT_FLAG);
    currentFlag |= _CRTDBG_LEAK_CHECK_DF;
    _CrtSetDbgFlag(currentFlag);
    _CrtSetBreakAlloc(-1L);
#endif

    scriptpath[0]=0; // ensure a null terminated string
    i = GetModuleFileName(NULL, scriptpath, sizeof(scriptpath));
    if ((i > 0) && (i < ((int)ARRAYSIZE(scriptpath)-(int)ARRAYSIZE(pathreplace)-(int)1))){

        scriptpath[ARRAYSIZE(scriptpath)-1] = 0; // limit strrchr search in case of errors (paranoia)

        (void)memmove(dllpath, scriptpath, sizeof scriptpath); // the buffer for the DLL path
        dlldir = _tcsrchr(dllpath, '\\'); // _tcsrchr -> strrchr / wcsrchr find the last backslash in path to get the file name
        dlldir = dlldir ? (dlldir + 1) : dllpath; // if find no backslash (unlikely), use the whole buffer
        emulate_perlexe = ((perlexe != 0) &&  (!_tcsnccmp(dlldir, perlexe, ARRAYSIZE(perlexe)))); //lint !e506 _tcsnccmp -> strncmp / wcsncmp

        TCHAR *rep = _tcsrchr(scriptpath, pathreplace[0]); // find the last delimiter in path
        if(!rep){
            (void)_ftprintf(stderr, _T("Failed to find '%C' in %s\n"), pathreplace[0], scriptpath);
            return 1; // ---> early return
        }
        if((pathreplace[0] == '.') && (pathreplace[1] == 0)){
            *rep = 0; // Perl script without extension, drop the '.'
        } else {
            (void)memmove(rep, pathreplace, sizeof pathreplace); // paste replacement
        }
    }
    else {
        (void)_ftprintf(stderr, _T("Path to %s is too long for my %i bytes buffer \n"), argv[0], sizeof(scriptpath));
        return 1; // ---> early return
    }

#ifdef DEBUGOUT
    _ftprintf(DEBUGOUT, _T("***** debug info *****\n%i argv parameters were passed to the exe:\n"), argc);
    for(i = 0; i < argc; i++){
        _ftprintf(DEBUGOUT, _T("%i:%s\n"), i, argv[i]);
    }
    if(!emulate_perlexe) _ftprintf(DEBUGOUT, _T("\nScript to be called: \"%s\"\n"), scriptpath);
    else                 _ftprintf(DEBUGOUT, _T("\nWe emulate the original perl.exe\n"));
#endif

#ifdef EXPLICITLINKING
    HINSTANCE hDLL;      // Handle to Perl DLL
    RunPerl_t RunPerl;   // Function pointer
    HANDLE hFind;        // for FindFirstFile()
    WIN32_FIND_DATA ffd; // for FindFirstFile()

    CASSERT(((sizeof dllsearch)+(sizeof scriptpath)) < (sizeof dllpath));
    // the dllpath buffer is longer than the scriptpath buffer => memmove can't overflow:
    (void)memmove(dlldir, dllsearch, sizeof dllsearch); // build search spec for the Perl DLL
    // remember that dlldir currently points to the last backslash of the path to the exe

#ifdef DEBUGOUT
    _ftprintf(DEBUGOUT, _T("DLL search spec:     \"%s\"\n"), dllpath);
#endif

    hFind = FindFirstFile(dllpath, &ffd);
    if(hFind == INVALID_HANDLE_VALUE){
        (void)_ftprintf(stderr, _T("Could not find %s\n"), dllpath);
        return 1; // ---> early return
    }
    // now we have the name of the DLL (without path) in ffd.cFileName

    // search again since our DLL search spec could contain again a backslash!
    dlldir = _tcsrchr(dllpath, '\\'); // find the last backslash in path to get the DLL directory
    dlldir = dlldir ? (dlldir + 1) : dllpath; // if find no backslash (unlikely), set to start of buffer
    // note: a trailing slash is needed for the (stupid) case of a directory with a trailing space!

    *dlldir = 0; // strip the file name spec from the path
    (void)SetDllDirectory(dllpath); // add the directory to the search path
    // as a positive side-effect, this removes the current directory from the search path

#ifdef DEBUGOUT
    _ftprintf(DEBUGOUT, _T("DLL name found:      \"%s\" (length = %i)\n"), ffd.cFileName, _tcslen(ffd.cFileName) ); // _tcslen -> stgrlen / wcslen
    _ftprintf(DEBUGOUT, _T("DLL search path set: \"%s\"\n"), dllpath);
#endif

    hDLL = LoadLibrary(ffd.cFileName); // searches first in the directory set by SetDllDirectory()
    if (!hDLL){
        (void)_ftprintf(stderr, _T("Failed to load Perl DLL \"%s\" code %li\n"), ffd.cFileName, GetLastError());
        (void)FindClose(hFind); // invalidates / frees ffd.cFileName
        return 1; // ---> early return
    }
    (void)FindClose(hFind); // invalidates / frees ffd.cFileName

    RunPerl = (RunPerl_t)GetProcAddress(hDLL, "RunPerl"); // this name works with ActiveState and Strawberry
    if (!RunPerl){
       (void)FreeLibrary(hDLL);
       (void)_ftprintf(stderr, _T("Failed to get RunPerl address in DLL. Check the DLL for name mangling.\n"));
       return 1; // ---> early return
    }
#endif // EXPLICITLINKING

#ifdef DEBUGOUT
    _ftprintf(DEBUGOUT, _T("***** end of debug info *****\n"));
#endif

    if(emulate_perlexe){

#ifdef UNICODE
        /// \hack - this is just for testing Perl / Windows unicode capabilities and problems!
        char **ourargv = (char **)malloc((argc+ARGS_ADDED+1) * sizeof(char**));
        if (!ourargv) { (void)_ftprintf(stderr, _T("Out of memory building new arg list\n")); return 1; } // ---> early return

        ourargv[0] = encode(argv[0], CP_UTF8); // keep filename, although it seems to be dropped by perl

        ourargv[1] = "-CA"; // -CA signals that @ARGV elements are encoded in UTF-8 -> https://perldoc.perl.org/perlrun.html
        // Check resulting flags in Perl with ${^UNICODE}

        for (i=1; i<argc; ++i) ourargv[i+ARGS_ADDED] = encode(argv[i], CP_UTF8);
        ourargv[argc+ARGS_ADDED] = NULL;

        int envc;
        for (envc = 0; env[envc]; envc++);
        char **ourenv = (char **)malloc((envc+1) * sizeof(char**));;

        for (i=0; i<envc; i++){
            ourenv[i] = encode(env[i], CP_UTF8); // CP_UTF8 or CP_ACP
//            printf("ourenv(%d): %s\n", i,ourenv[i]); // debug hack to verify translation
        }
        ourenv[envc] = NULL;

#ifdef DEBUGOUT
//        (void)_ftprintf(DEBUGOUT, _T("environ is %x; env is %x; ourenv is %x\n"), (int)environ, (int)env, (int)ourenv);
//        (void)_ftprintf(DEBUGOUT, _T("argc is %d\n"), argc+ARGS_ADDED);
//        printf("argv4u: %s\n", ourargv[4]); // debug hack
        fflush(stderr); // needed before _setmode
        _setmode(_fileno(stderr), _O_TEXT); // else Perl errors won't be displayed!
#endif
        // If we pass NULL as env pointer, Perl uses the system provided ANSI environ pointer
        // As far as I see, Unicode environment is mix: names are our UTF8 strings, values are ANSI strings!
        i = RunPerl(argc+ARGS_ADDED, ourargv, ourenv); // emulate the standard perl.exe
#else // UNICODE
        i = RunPerl(argc, argv, env); // emulate the standard perl.exe
#endif // UNICODE

#ifdef EXPLICITLINKING
        if(hDLL) (void)FreeLibrary(hDLL);
#endif
    } else {
        // generate "our" argument list with additional entries and terminating NULL pointer
        char **ourargv = (char **)malloc((argc+ARGS_ADDED+1) * sizeof(char**));
        if (!ourargv) { (void)_ftprintf(stderr, _T("Out of memory building new arg list\n")); return 1; } // ---> early return

#ifdef UNICODE
        ourargv[0] = encode(argv[0], CP_ACP); // keep filename, although it seems to be dropped by perl

        // pass script path to perl interpreter, encoded in the system code page,
        // just as it is passed via main() rather than wmain(). Perl implementations
        // open the script path with 8-bit i/o calls rather than the wide calls,
        // regardless of the setting of -Ci.
        ourargv[1] = encode(scriptpath, CP_ACP);

        // copy the remaining user provided arguments encoded in UTF-8.
        for (i=1; i<argc; ++i) ourargv[i+ARGS_ADDED] = encode(argv[i], CP_UTF8);
        ourargv[argc+ARGS_ADDED] = NULL;

        // make new env encoded in the system code page, just as it is passed via main().
        int envc = 0;
        for (; env[envc]; envc++);
        char **ourenv = (char **)malloc((envc+1) * sizeof(char**));;
        for (i=0; i<envc; i++) ourenv[i] = encode(env[i], CP_ACP);
        ourenv[envc] = NULL;

        i = RunPerl(argc+ARGS_ADDED, ourargv, ourenv);
#else // UNICODE
        ourargv[0] = argv[0]; // keep filename, although it seems to be dropped by perl
        ourargv[1] = scriptpath; // pass script path to Perl interpreter

        // copy the remaining user provided arguments and the terminating NULL pointer
        for (i=1; i<=argc; ++i) {
            ourargv[i+ARGS_ADDED] = argv[i];
        }

        i = RunPerl(argc+ARGS_ADDED, ourargv, env);
#endif // UNICODE

#ifdef EXPLICITLINKING
        if(hDLL) (void)FreeLibrary(hDLL);
#endif
        free(ourargv);
    }
    fprintf(stdout, "\nend of perl.exe - why does 'return i' take so long?\n");
    return i; // pass the return code from the Perl interpreter like original perl.exe does
}

